"""handle geo-referenced raster images"""
from affine import Affine
from numpy import ndarray as np_ndarray

from samgis_core.utilities.type_hints import ListFloat, DictStrInt, TupleFloat
from samgis_core import app_logger


def load_affine_transformation_from_matrix(matrix_source_coefficients: ListFloat) -> Affine:
    """
    Wrapper for rasterio.Affine.from_gdal() method. Useful to read WorldFiles (TFW, PGW)

    Args:
        matrix_source_coefficients: 6 floats ordered by GDAL.

    Returns:
        Affine transform

    """

    if len(matrix_source_coefficients) != 6:
        raise ValueError(f"Expected 6 coefficients, found {len(matrix_source_coefficients)}; "
                         f"argument type: {type(matrix_source_coefficients)}.")

    try:
        a, d, b, e, c, f = (float(x) for x in matrix_source_coefficients)
        center = tuple.__new__(Affine, [a, b, c, d, e, f, 0.0, 0.0, 1.0])
        return center * Affine.translation(-0.5, -0.5)
    except Exception as e:
        app_logger.exception(f"exception:{e}, check updates on https://github.com/rasterio/affine",
                             extra=e,
                             stack_info=True, exc_info=True)
        raise e


def get_vectorized_raster_as_geojson(mask: np_ndarray, transform: TupleFloat) -> DictStrInt:
    """
        Get shapes and values of connected regions in a dataset or array

        Args:
            mask: numpy mask
            transform: tuple of float to transform into an Affine transform

        Returns:
            dict containing the output geojson and the geojson shapes number

    """
    try:
        from rasterio.features import shapes
        from geopandas import GeoDataFrame

        app_logger.debug(f"matrix to consume with rasterio.shapes: {type(transform)}, {transform}.")

        # old value for mask => band != 0
        shapes_generator = ({
            'properties': {'raster_val': v}, 'geometry': s}
            for i, (s, v)
            # instead of `enumerate(shapes(mask, mask=(band != 0), transform=rio_src.transform))`
            # use mask=None to avoid using source
            in enumerate(shapes(mask, mask=None, transform=transform))
        )
        app_logger.info("created shapes_generator, transform it to a polygon list...")
        shapes_list = list(shapes_generator)
        app_logger.info(f"created {len(shapes_list)} polygons.")
        gpd_polygonized_raster = GeoDataFrame.from_features(shapes_list, crs="EPSG:3857")
        app_logger.info("created a GeoDataFrame, export to geojson...")
        geojson = gpd_polygonized_raster.to_json(to_wgs84=True)
        app_logger.info("created geojson, preparing API response...")
        return {
            "geojson": geojson,
            "n_shapes_geojson": len(shapes_list)
        }
    except Exception as e_shape_band:
        try:
            app_logger.error(f"mask type:{type(mask)}.")
            app_logger.error(f"transform type:{type(transform)}, {transform}.")
            app_logger.error(f"mask shape:{mask.shape}, dtype:{mask.dtype}.")
        except Exception as e_shape_dtype:
            app_logger.exception(f"mask shape or dtype not found:{e_shape_dtype}.", exc_info=True)
        app_logger.exception(f"e_shape_band:{e_shape_band}.", exc_info=True)
        raise e_shape_band



"""web helper functions"""
from typing import Dict

from xyzservices import providers, TileProvider
from samgis_core import app_logger
from samgis_web.io_package.coordinates_pixel_conversion import get_latlng_to_pixel_coordinates
from samgis_web.utilities.constants import COMPLETE_URL_TILES_MAPBOX, COMPLETE_URL_TILES_NEXTZEN
from samgis_web.utilities.type_hints import ApiRequestBody, XYZTerrainProvidersNames, XYZDefaultProvidersNames


def get_parsed_bbox_points_with_dictlist_prompt(request_input: ApiRequestBody | str) -> Dict:
    """
    Parse the raw input request into bbox, prompt and zoom.
    If the request is a string this function use 'model_validate_json()' method from instance Pydantic BaseModel.

    Args:
        request_input: input dict

    Returns:
        dict with bounding box, prompt and zoom

    """
    app_logger.info(f"try to parsing input request: {type(request_input)}, {request_input}...")
    if isinstance(request_input, str):
        app_logger.info(f"string/json input, parsing it to {type(ApiRequestBody)}...")
        request_input = ApiRequestBody.model_validate_json(request_input)
        app_logger.info(f"parsed input, now of type {type(request_input)}...")

    bbox = request_input.bbox
    app_logger.debug(f"request bbox: {type(bbox)}, value:{bbox}.")
    ne = bbox.ne
    sw = bbox.sw
    app_logger.debug(f"request ne: {type(ne)}, value:{ne}.")
    app_logger.debug(f"request sw: {type(sw)}, value:{sw}.")
    ne_latlng = [float(ne.lat), float(ne.lng)]
    sw_latlng = [float(sw.lat), float(sw.lng)]
    new_zoom = int(request_input.zoom)
    new_prompt_list = _get_parsed_prompt_list(ne, sw, new_zoom, request_input.prompt)

    app_logger.debug(f"bbox => {bbox}.")
    app_logger.debug(f'request_input-prompt updated => {new_prompt_list}.')

    app_logger.info("unpacking elaborated request...")
    return {
        "bbox": [ne_latlng, sw_latlng],
        "prompt": new_prompt_list,
        "zoom": new_zoom,
        "source": get_source_tile(request_input.source_type),
        "source_name": get_source_name(request_input.source_type)
    }


def _get_parsed_prompt_list(bbox_ne, bbox_sw, zoom, prompt_list):
    new_prompt_list = []
    for prompt in prompt_list:
        app_logger.debug(f"current prompt: {type(prompt)}, value:{prompt}.")
        new_prompt = {"type": prompt.type.value}
        if prompt.type == "point":
            new_prompt_data = _get_new_prompt_data_point(bbox_ne, bbox_sw, prompt, zoom)
            new_prompt["label"] = prompt.label.value
        elif prompt.type == "rectangle":
            new_prompt_data = _get_new_prompt_data_rectangle(bbox_ne, bbox_sw, prompt, zoom)
        else:
            msg = "Valid prompt type: 'point' or 'rectangle', not '{}'. Check ApiRequestBody parsing/validation."
            raise TypeError(msg.format(prompt.type))
        app_logger.debug(f"new_prompt_data: {type(new_prompt_data)}, value:{new_prompt_data}.")
        new_prompt["data"] = new_prompt_data
        new_prompt_list.append(new_prompt)
    return new_prompt_list


def _get_new_prompt_data_point(bbox_ne, bbox_sw, prompt, zoom):
    current_point = get_latlng_to_pixel_coordinates(bbox_ne, bbox_sw, prompt.data, zoom, prompt.type)
    app_logger.debug(f"current prompt: {type(current_point)}, value:{current_point}, label: {prompt.label}.")
    return [current_point['x'], current_point['y']]


def _get_new_prompt_data_rectangle(bbox_ne, bbox_sw, prompt, zoom):
    current_point_ne = get_latlng_to_pixel_coordinates(bbox_ne, bbox_sw, prompt.data.ne, zoom, prompt.type)
    app_logger.debug(
        f"rectangle:: current_point_ne prompt: {type(current_point_ne)}, value:{current_point_ne}.")
    current_point_sw = get_latlng_to_pixel_coordinates(bbox_ne, bbox_sw, prompt.data.sw, zoom, prompt.type)
    app_logger.debug(
        f"rectangle:: current_point_sw prompt: {type(current_point_sw)}, value:{current_point_sw}.")
    # correct order for rectangle prompt
    return [
        current_point_sw["x"],
        current_point_ne["y"],
        current_point_ne["x"],
        current_point_sw["y"]
    ]


mapbox_terrain_rgb = TileProvider(
    name=XYZTerrainProvidersNames.MAPBOX_TERRAIN_TILES_NAME,
    url=COMPLETE_URL_TILES_MAPBOX,
    attribution=""
)
nextzen_terrain_rgb = TileProvider(
    name=XYZTerrainProvidersNames.NEXTZEN_TERRAIN_TILES_NAME,
    url=COMPLETE_URL_TILES_NEXTZEN,
    attribution=""
)


def get_source_tile(source_type: str) -> TileProvider:
    """
    Return `TileProvider` instance based on the name query.
    See examples in xyzservices.lib.py, query_name() for details.

    Args:
        source_type: source name to use in query_name()

    Returns:
        a TileProvider instance
    """
    try:
        match source_type.lower():
            case XYZDefaultProvidersNames.DEFAULT_TILES_NAME_SHORT:
                return providers.query_name(XYZDefaultProvidersNames.DEFAULT_TILES_NAME)
            case XYZTerrainProvidersNames.MAPBOX_TERRAIN_TILES_NAME:
                app_logger.info(f"mapbox_terrain_rgb:{mapbox_terrain_rgb.name}.")
                return mapbox_terrain_rgb
            case XYZTerrainProvidersNames.NEXTZEN_TERRAIN_TILES_NAME:
                app_logger.info(f"nextzen_terrain_rgb:{nextzen_terrain_rgb.name}.")
                return nextzen_terrain_rgb
            case _:
                return providers.query_name(source_type)
    except ValueError as ve:
        from pydantic_core import ValidationError

        app_logger.error(f"ve:{ve}.")
        ve_title = str(ve)
        raise ValidationError.from_exception_data(title=ve_title, line_errors=[])


get_url_tile = get_source_tile


def check_source_type_is_terrain(source: str | TileProvider) -> bool:
    """Check if the given source string or TileProvider is a terrain one, see

     - Digital Elevation Model ([DEM](https://www.usgs.gov/faqs/what-a-digital-elevation-model-dem))
     - Digital Terrain Model ([DTM](https://www.usgs.gov/news/how-create-quality-digital-terrain-model))

     Args:
        source: source string or TileProvider

     Returns:
        a boolean (True if is a DEM/DTM/terrain TileProvider, False otherwise)

    """
    return isinstance(source, TileProvider) and source.name in list(XYZTerrainProvidersNames)


def get_source_name(source: str | TileProvider) -> str | bool:
    """
    Return `TileProvider.name` string based on the output from query_name().
    See examples in xyzservices.lib.py, query_name() for details.

    Args:
        source: source string or TileProvider

    Returns:
        a string name representing the input TileProvider or source string

    """
    try:
        match source.lower():
            case XYZDefaultProvidersNames.DEFAULT_TILES_NAME_SHORT:
                source_output = providers.query_name(XYZDefaultProvidersNames.DEFAULT_TILES_NAME)
            case _:
                source_output = providers.query_name(source)
        if isinstance(source_output, str):
            return source_output
        try:
            source_dict = dict(source_output)
            app_logger.info(f"source_dict:{type(source_dict)}, {'name' in source_dict}, source_dict:{source_dict}.")
            return source_dict["name"]
        except KeyError as ke:
            app_logger.error(f"ke:{ke}.")
    except ValueError as ve:
        app_logger.info(f"source name::{source}, ve:{ve}.")
    app_logger.info(f"source name::{source}.")

    return False



"""custom type hints"""
from enum import IntEnum, Enum
from typing import TypedDict

from affine import Affine
from numpy import ndarray
from pydantic import BaseModel

from samgis_core.utilities.type_hints import StrEnum


tuple_ndarray_transform = tuple[ndarray, Affine]


class WorldFile(StrEnum):
    """Default xyz provider names"""
    pgw = "pgw"
    tfw = "tfw"


class XYZDefaultProvidersNames(StrEnum):
    """Default xyz provider names"""
    DEFAULT_TILES_NAME_SHORT = "openstreetmap"
    DEFAULT_TILES_NAME = "openstreetmap.mapnik"


class XYZTerrainProvidersNames(StrEnum):
    """Custom xyz provider names for digital elevation models"""
    MAPBOX_TERRAIN_TILES_NAME = "mapbox.terrain-rgb"
    NEXTZEN_TERRAIN_TILES_NAME = "nextzen.terrarium"


class LatLngDict(BaseModel):
    """Generic geographic latitude-longitude type"""
    lat: float
    lng: float


class ContentTypes(str, Enum):
    """Segment Anything: validation point prompt type"""
    APPLICATION_JSON = "application/json"
    TEXT_PLAIN = "text/plain"
    TEXT_HTML = "text/html"


class PromptPointType(str, Enum):
    """Segment Anything: validation point prompt type"""
    point = "point"


class PromptRectangleType(str, Enum):
    """Segment Anything: validation rectangle prompt type"""
    rectangle = "rectangle"


class PromptLabel(IntEnum):
    """Valid prompt label type"""
    EXCLUDE = 0
    INCLUDE = 1


class ImagePixelCoordinates(TypedDict):
    """Image pixel coordinates type"""
    x: int
    y: int


class RawBBox(BaseModel):
    """Input lambda bbox request type (not yet parsed)"""
    ne: LatLngDict
    sw: LatLngDict


class RawPromptPoint(BaseModel):
    """Input lambda prompt request of type 'PromptPointType' - point (not yet parsed)"""
    type: PromptPointType
    data: LatLngDict
    label: PromptLabel


class RawPromptRectangle(BaseModel):
    """Input lambda prompt request of type 'PromptRectangleType' - rectangle (not yet parsed)"""
    type: PromptRectangleType
    data: RawBBox

    def get_type_str(self):
        return self.type


class ApiRequestBody(BaseModel):
    """Input request validator type (not yet parsed)"""
    id: str = ""
    bbox: RawBBox
    prompt: list[RawPromptPoint | RawPromptRectangle]
    zoom: int | float
    source_type: str = "OpenStreetMap.Mapnik"
    debug: bool = False


class ApiResponseBodyFailure(BaseModel):
    """SamGIS API Response; handle only case of failure"""
    duration_run: float
    message: str
    request_id: str


class ApiResponseBodySuccess(ApiResponseBodyFailure):
    """SamGIS API Response; handle both case of success and failure"""
    n_predictions: int
    geojson: str
    n_shapes_geojson: int



from typing import Callable

import gradio as gr

from samgis_core.utilities.type_hints import ListStr, DictStrInt


def get_example_complete(example_text, example_body, key_text: str = "string_prompt") -> str:
    """
    return an example dict, as in {"string_prompt": "some text..."}

    Args:
        example_text: example text
        example_body: a request example body placeholder
        key_text: dict key for text string

    Returns:
        an example formatted json string request

    """
    import json
    example_dict = dict(**example_body)
    if key_text in example_dict:
        example_dict[key_text] = example_text
    return json.dumps(example_dict)


def get_gradio_interface_geojson(
        fn_inference: Callable, markdown_text: str, examples_text_list_text: ListStr, example_body: DictStrInt):
    """
    Return a Gradio interface composed by a row, two column (one text input and one text output)

    Args:
        fn_inference: function used within the Gradio interface
        markdown_text: Markdown description
        examples_text_list_text: a list of example formatted json string requests
        example_body: a request example body placeholder

    Returns:
        a Gradio interface

    """
    with gr.Blocks() as gradio_app:
        gr.Markdown(markdown_text)

        with gr.Row():
            with gr.Column():
                text_input = gr.Textbox(lines=1, placeholder=None, label="Payload input")
                btn = gr.Button(value="Submit")
            with gr.Column():
                text_output = gr.Textbox(lines=1, placeholder=None, label="Geojson Output")

        gr.Examples(
            examples=[
                get_example_complete(example, example_body) for example in examples_text_list_text
            ],
            inputs=[text_input],
        )
        btn.click(
            fn_inference,
            inputs=[text_input],
            outputs=[text_output]
        )
    return gradio_app
